:- dynamic(v/1).
:- dynamic(j/1).
:- dynamic(c/1).
:- dynamic(vid/1).
:- dynamic(xhalv/1).

:- op(830,fx,xhalv).

:- ensure_loaded('util.pl').
:- ensure_loaded('valid.pl').
:- ensure_loaded('invalid.pl').
:- ensure_loaded('join.pl').

flag_vreport(4).

:- add_nb_set(vgenid/1,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% COMPLETENESS                                                             %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- MISSING RULES ------------------------------------------------------------%
%------------------------------------------------------------------------------%	

vcomplete(Schema) :- 
  vgenid(_),		% Non-backtrackable predicate
  xschm Schema: Attrs ==> _, 
  generate(Attrs,Values),
  cover(Schema, Attrs, Values), 
  fail.
vcomplete(_) :- 
  findall(C,c(C),[]),!.
vcomplete(_) :-  
  retractall(c(_)),!.

generate([],[]).
generate([A|ARest], State) :- 
  alsv_domain(A, Domain,_), 
  vnormalize(A,Domain, NDomain), 
  member(V,NDomain),			% if general, put it in brackets
  generate(ARest, AState), 
  append(V, AState, State).

cover(Schema, Attrs, Values) :-
  assertState(Attrs,Values),
  xrule(Schema/_: Facts ==> _), 
  covered(Facts), 
  retractState(Attrs,Values),!.
cover(Schema, Attrs, Values) :- 
  retractState(Attrs,Values), 
  generateRule(Attrs,Values), 
  vconcat_ws(', ', Attrs, AString),
  vconcat_ws(', ', Values, VString),
  vconcat(['\'In ', Schema,' uncover state: (',AString,') = (',VString,')\''],RString), 
  vreport(1,[RString]), nl .

covered([]) :- !.
covered([Fact|Facts]) :- 
  vcondition(Fact,VFact),
  alsv_valid(VFact,vstate), 
  covered(Facts).
  
generateRule([],[]) :- !.
generateRule([A|Attrs],[V|Values]) :- 
  c(A in ValSet),
  flatten([V], NV),
  appendDistinct(NV, ValSet,NewValSet),
  retract(c(A in ValSet)),
  assert(c(A in NewValSet)),
  generateRule(Attrs,Values),!.
generateRule([A|Attrs],[V|Values]) :-
  flatten([V], FV),
  assert(c(A in FV)),
  generateRule(Attrs,Values). 
   
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% DETERMINISM                                                              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- CONTRADICTION ------------------------------------------------------------%
%------------------------------------------------------------------------------%
vcontradict(Schema) :- 
  vgenid(_), % Non-backtrackable predicate
  xschm Schema: Attrs ==> _, 
  generate(Attrs,Values), 
  contr(Schema, Attrs, Values),
  fail.
vcontradict(_). 

contr(Schema, Attrs, Values) :- 
  assertState(Attrs,Values),
  xrule(Schema/Id: Facts ==> Decs), 
  xrule(Schema/Id2: Facts2 ==> Decs2),
  Id \= Id2, 
  covered(Facts), 
  covered(Facts2), 
  conflict(Decs, Decs2),
  retractState(Attrs,Values),
  vconcat_ws(', ', Attrs, AString),
  vconcat_ws(', ', Values, VString),
  vconcat(['\'In ', Schema,' rule: ',Id,' conflicts with rule ',Id2,' for state: (',AString,') = (',VString,')\''],RString), 
  vreport(0,[Id,RString]),
  vreport(0,[Id2,RString]),!.
contr(_, Attrs, Values) :- 
  retractState(Attrs,Values).
  
conflict(Decs1, Decs2):-
  split_decision(Decs1,[D1,_,_,_]), 
  split_decision(Decs2,[D2,_,_,_]),
  conflict_dec_attributes(D1,D2).
  
conflict_dec_attributes([],[]) :- !.
conflict_dec_attributes([D1|Ds1],[D2|Ds2]) :-  
  conflict_item(D1,D2),
  conflict_dec_attributes(Ds1,Ds2).  

conflict_item(D1,D2) :-
  retractall(v(_)),
  assert(v(D2)), 
  invalid(D1).

%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% MIN. NUMBER OF RULES                                                     %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- SUBSUMPTION --------------------------------------------------------------%
%------------------------------------------------------------------------------%
vsubsume(Schema) :- 
  vgenid(_), % Non-backtrackable predicate
  xrule(Schema/Id1:Cs1==>Ds1),
  xrule(Schema/Id2:Cs2==>Ds2),
  Id1 \= Id2, 
  subsumed_cond(Cs1,Cs2),    
  subsumed_dec(Ds1, Ds2),  
  vconcat(['\'In ', Schema,' rule: ',Id2,' subsumes rule ',Id1,'\''],RString), 
  vreport(0,[Id1,RString]),
  vreport(0,[Id2,RString]),
  fail.
vsubsume(_) :-  
  retractall(v(_)).

subsumed_cond([],[]) :- !. 
subsumed_cond([C1|Cs1],[C2|Cs2]) :-
  vcondition(C1, VC1),
  vcondition(C2, VC2),  
  sub(VC1,VC2),
  subsumed_cond(Cs1,Cs2).  
  
sub(C1,C2) :-
  retractall(v(_)),
  assert(v(C2)), 
  valid(C1).

subsumed_dec(Ds1, Ds2) :- 
  split_decision(Ds1,[D1,_,_,_]), 
  split_decision(Ds2,[D2,_,_,_]), 
  subsumed_dec_attributes(D1,D2).

subsumed_dec_attributes([],[]) :- !.
subsumed_dec_attributes([D1|Ds1],[D2|Ds2]) :- 
  sub(D1,D2),
  subsumed_dec_attributes(Ds1,Ds2).  

%------------------------------------------------------------------------------%
%--- REDUCTION ----------------------------------------------------------------%
%------------------------------------------------------------------------------%
vreduce(Schema):-
  vgenid(_), % Non-backtrackable predicate
  xrule(Schema/Id1:Cs1==>Ds1),
  xrule(Schema/Id2:Cs2==>Ds2),
  Id1 \= Id2, 
  split_decision(Ds1,[D1,_,_,_]), 
  split_decision(Ds2,[D2,_,_,_]), 
  D1 == D2,
  join_conds(Cs1,Cs2,Cond),
  neqset(Cond,[]),
  vconcat(['\'In ', Schema,' rule: ',Id1,' can be joined with rule ',Id2,'\''],RString), 
  vreport(0,[Id1,RString]),
  vreport(0,[Id2,RString]),
  fail.
vreduce(_) :- 
  retractall(j(_)).

join_conds([],[],[]) :- !.
join_conds([C1|Cs1],[C2|Cs2],[C|Cond]) :- 
  vcondition(C1, VC1),
  vcondition(C2, VC2), 
  join_cond(VC1, VC2, C),  
  join_conds(Cs1,Cs2,Cond).
  
join_cond(C1,C2,C) :-
  retractall(j(_)),
  assert(j(C2)),
  join(C1, C).
  
  